@@ -1,9 +1,9 @@
 ;;; window-layout.el --- window layout manager
 
-;; Copyright (C) 2010, 2011, 2012, 2013  SAKURAI Masashi
+;; Copyright (C) 2010-2017  SAKURAI Masashi
 
 ;; Author: SAKURAI Masashi <m.sakurai atmark kiwanami.net>
-;; Version: 1.2
+;; Version: 1.4
 ;; Keywords: window, layout
 
 ;; This program is free software; you can redistribute it and/or modify
@@ -35,11 +35,11 @@
 ;;            (- (:upper-max-size 15) 
 ;;               summary 
 ;;               message))
-;;        '((:name 'folder 
+;;        '((:name folder 
 ;;           :buffer "folder buffer")
-;;          (:name 'summary
+;;          (:name summary
 ;;           :buffer "summary buffer")
-;;          (:name 'message
+;;          (:name message
 ;;           :buffer "message buffer")
 ;;         )))
 
@@ -93,7 +93,7 @@
 ;; * Window options:
 
 ;;   :name  [*] the window name.
-;;   :buffer  a buffer name or a buffer object to show the window. If nil or omitted, the current buffer remains.
+;;   :buffer  a buffer name or a buffer object to show the window. If nil or omitted, the current buffer remains. If symbol, it is evaluated as a global variable.
 ;;   :default-hide  (t/nil) if t, the window is hided initially. (default: nil)
 ;;   :fix-size  (t/nil) if t, when the windows are laid out again, the window size is remained. (default: nil)
 
@@ -123,7 +123,7 @@
 (eval-when-compile (require 'cl))
 
 (defmacro wlf:aif (test-form then-form &rest else-forms)
-  (declare (debug ("test-form" form "then-form" form &rest form))
+  (declare (debug (form form &rest form))
            (indent 2))
   `(let ((it ,test-form))
      (if it ,then-form ,@else-forms)))
@@ -138,7 +138,7 @@
              (let ((it ,sym)) ,@(cdr cl1))
            (wlf:acond ,@(cdr clauses)))))))
 
-(defun wlf:window-first-line-point (window)
+(defun wlf:current-first-line-point (window)
   "[internal] return the point at the beginning of the first line
 of the WINDOW."
   (when (windowp window)
@@ -148,25 +148,32 @@ of the WINDOW."
           (move-to-window-line 0)
           (point))))))
 
-;;; Window-set management structure
-;; recipe      : an input recipe object.
-;; winfo-list  : a list of window management structures.
-;; wholep      : if non nil, this function uses whole frame window.
-;; layout-hook : if doing layout windows, these hooks are called. 
-;;               The hook function has one argument: wset object.
+(defstruct wlf:wset
+"Window-set management structure
 
-(defstruct wlf:wset recipe winfo-list wholep layout-hook)
+recipe      : an input recipe object.
+winfo-list  : a list of window management structures.
+wholep      : if non nil, this function uses whole frame window.
+layout-hook : if doing layout windows, these hooks are called. 
+              The hook function has one argument: wset object."
+recipe winfo-list wholep layout-hook)
 
-;;; Window management structure
-;; name      : a symbol of the window name.
-;; options   : an option plist given by the recipe.
-;; shown     : 'show/'hide. if 'hide, the window is not displayed.
-;; window    : a window object.
-;; vertical  : if the window is split vertically, the value is t.
-;; last-size : if the window is alive, the window size is saved before laying out.
-;; edges     : a list of window edges returned by `window-edges'.
+(defstruct wlf:window
+"Window management structure
 
-(defstruct wlf:window name options shown window vertical last-size edges)
+name      : a symbol of the window name.
+options   : an option plist given by the recipe.
+shown     : 'show/'hide. if 'hide, the window is not displayed.
+window    : a window object.
+vertical  : if the window is split vertically, the value is t.
+last-size : if the window is alive, the window size is saved before laying out.
+edges     : a list of window edges returned by `window-edges'.
+first-line-point : the cursor position of the window.
+window-point     : the scroll position of the window.
+(If more than 2 windows display the same buffer, 
+the values of first-line-point and window-point may 
+be different between windows.)"
+name options shown window vertical last-size edges first-line-point window-point)
 
 (defun wlf:window-shown-set (winfo i)
   "[internal] translate the argument: nil -> 'hide / t -> 'show"
@@ -194,7 +201,7 @@ in current frame."
            (and swin 
                 (destructuring-bind
                     (sl st sr sb) (window-edges swin)
-                  (if (and (equal left sl) (equal top st)) t
+                  (if (and (equal left sl) (< (abs (- top st)) 3)) t
                     (message "OLD:%S  NEW:%S"
                              (wlf:window-edges winfo) (window-edges swin)) nil))
                 swin)))))
@@ -340,26 +347,34 @@ start dividing."
   (if (not (wlf:window-shown-p winfo))
       (delete-window (selected-window))
     (let ((buffer (wlf:aif (wlf:window-option-get winfo :buffer)
-                      (get-buffer it))))
+                      (get-buffer 
+                       (if (symbolp it) 
+                           (symbol-value it) it)))))
       (when (buffer-live-p buffer)
         (set-window-buffer (selected-window) buffer)))))
 
 (defun wlf:set-window-points (winfo)
-  "[internal] Set window points recorded in WINFO."
+  "[internal] Set the scroll position and cursor one which are recorded in WINFO."
   (let ((window (wlf:window-window winfo)))
     (when (window-live-p window)
       (with-selected-window window
-        (wlf:aif (wlf:window-option-get winfo :window-first-line-point)
+        (wlf:aif (wlf:window-first-line-point winfo)
             (progn
+              ;;(message "WLF:set-window-points: [%s] first-line: %i" (wlf:window-name winfo) it)
               (goto-char it)
               (recenter 0)))
-        (wlf:aif (wlf:window-option-get winfo :window-point)
+        ;; Go to previously saved window point if point is shown or
+        ;; end of buffer is shown (therefore the point is shown).  In
+        ;; the latter case, window point (`it') is also at the EOB
+        ;; (therefore `<' check does not work).
+        (wlf:aif (wlf:window-window-point winfo)
             (progn
-              ;; window-point may not be in visible range.  In that
-              ;; case, do not set point.
               (let ((win-last-point (window-end nil t)))
-                (when (<= it win-last-point)
-                  (goto-char it)))))))))
+                ;;(message "WLF:set-window-points: [%s] window-point: %i  last-point: %i" (wlf:window-name winfo) it win-last-point)
+                (goto-char (if (or (< it win-last-point)
+                                   (= (point-max) win-last-point))
+                               it
+                             (1- win-last-point))))))))))
 
 (defun wlf:collect-window-edges (winfo-list)
   "[internal] At the end of window laying out, this function is
@@ -488,12 +503,11 @@ The saved points are used in `wlf:apply-winfo'."
                   (setf (wlf:window-last-size winfo)
                         (and win
                              (wlf:max-window-size-p winfo)
-                             (wlf:window-size winfo)))
-                  (plist-put (wlf:window-options winfo)
-                             :window-point (and win (window-point win)))
-                  (plist-put (wlf:window-options winfo)
-                             :window-first-line-point
-                             (wlf:window-first-line-point win))))))
+                             (wlf:window-size winfo))
+                        (wlf:window-window-point winfo)
+                        (and win (window-point win))
+                        (wlf:window-first-line-point winfo)
+                        (wlf:current-first-line-point win))))))
   (set-frame-parameter (selected-frame) 'wlf:recipe recipe))
 
 
@@ -516,7 +530,8 @@ layout. See the comment of `wlf:layout' function for arguments."
                  :wholep (not subwindow-p)))
 
 (defmacro wlf:with-wset (wset &rest body)
-  (declare (debug ("wset" sexp &rest form))
+  "Define local variables: recipe, winfo-list, wholep, layout-hook."
+  (declare (debug (symbolp &rest form))
            (indent 1))
   `(let* 
        ((recipe (wlf:wset-recipe wset))
@@ -653,17 +668,18 @@ name or object to show in the window."
   (let* ((winfo 
           (wlf:get-winfo 
            winfo-name (wlf:wset-winfo-list wset)))
-         (window (wlf:window-live-window winfo)))
+         (window (wlf:window-live-window winfo))
+         (curwin (selected-window)))
     (unless (buffer-live-p buf)
       (error "Buffer is dead. at wlf:set-buffer. (%s)" winfo-name))
     (plist-put (wlf:window-options winfo) :buffer buf)
-    (plist-put (wlf:window-options winfo) :window-point (window-point window))
-    (plist-put (wlf:window-options winfo)
-               :window-first-line-point (wlf:window-first-line-point window))
     (when (and window (not (eql (get-buffer buf) (window-buffer window))))
       (set-window-buffer window buf))
-    (when selectp 
+    (cond
+     (selectp 
       (select-window window))
+     ((active-minibuffer-window)
+      (select-window (minibuffer-window))))
     window))
 
 (defun wlf:get-buffer (wset winfo-name)
@@ -715,7 +731,7 @@ of a window name and a buffer object (or buffer name)."
   "[internal] Return a shallow copied window object."
   (make-wlf:window
    :name (wlf:window-name winfo)
-   :options (wlf:window-options winfo)
+   :options (copy-sequence (wlf:window-options winfo))
    :shown (wlf:window-shown winfo)
    :window (wlf:window-window winfo)
    :vertical (wlf:window-vertical winfo)
@@ -815,7 +831,17 @@ It returns WINDOW by given name."
                  (t (push sw windows))))
               finally return t)))))
 
+(defun wlf:wset-clear-window-points (wset)
+  "Clear the last preserved window-point and first-line-point slots for all windows.
+If users want change the window layout and re-use `wlf:wset' instance,
+  this function should be called to forget wrong window-positions."
+  (loop for winfo in (wlf:wset-winfo-list wset)
+        do
+        (setf (wlf:window-window-point winfo) nil
+              (wlf:window-first-line-point winfo) nil)))
+
 (defun wlf:wset-fix-windows (wset)
+  "Update window object instances with the current window configuration."
   (loop for winfo in (wlf:wset-winfo-list wset)
         for win = (wlf:window-window winfo)
         do
